#include <vector>
#include <algorithm>
#include <cmath>
#include "Image.h"
#include "Point.h"

Image::Image() {
    npx = npy = npix = 0;
    data = 0;
}

Image::Image(double* array, int n, int m) {
    data = new double[n*m];
    for (int j = 0; j < n; j++) {
	for (int i = 0; i < m; i++) {
	    data[i+j*m] = array[i+j*m];
	} 
    }
    npx = m;
    npy = n;
    npix = n * m;
}

Image::Image(const Image &img) {
    npx = img.getNpx();
    npy = img.getNpy();
    npix = npx * npy;
    data = new double[npix];
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] = img.getValue(i, j);
	}
    }
}

Image::Image(int npx, int npy) {
    this->data = new double[npx*npy];
    this->npx = npx;
    this->npy = npy;
    this->npix = npx * npy;
    for (int i = 0; i < npix; i++) {
	data[i] = 0.0;
    }
}

Image::~Image() {
    delete [] data;
}

int Image::getNpx() const {
    return npx;
}

int Image::getNpy() const {
    return npy;
}

int Image::getNpix() const {
    return npix;
}

void Image::getNaxis(int *npx, int *npy) {
    *npx = this->npx;
    *npy = this->npy;
}

double Image::getValue(int x, int y) const {
    return data[x+npx*y];
}

void Image::setValue(int x, int y, double v) {
    data[x+npx*y] = v;
}

void Image::getData(float *oarray, int n) {
    for (int j = 0; j < npx; j++) {
	for (int i = 0; i < npy; i++) {
	    oarray[j+i*npx] = this->data[j+i*npx];
	} 
    }
}

Image Image::crop(int xmin, int ymin, int npx, int npy) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(xmin+i, ymin+j));
	}
    }
    return out;
}

Image& Image::operator=(const Image &img) {
    npx = img.getNpx();
    npy = img.getNpy();
    npix = npx * npy;
    data = new double[npix];
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] = img.getValue(i, j);
	}
    }
}

Image& Image::operator+=(double a) {
    for (int i = 0; i < npix; i++) {
	this->data[i] += a;
    }
    return *this;
}

Image& Image::operator+=(const Image& a) {
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] += a.getValue(i,j);
	}
    }
    return *this;
}

Image Image::operator+(double a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) + a);
	}
    }
    return out;
}

Image Image::operator+(const Image& a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) + a.getValue(i,j));
	}
    }
    return out;
}

Image& Image::operator-=(double a) {
    for (int i = 0; i < npix; i++) {
	this->data[i] -= a;
    }
    return *this;
}

Image& Image::operator-=(const Image& a) {
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] -= a.getValue(i,j);
	}
    }
    return *this;
}

Image Image::operator-(double a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) - a);
	}
    }
    return out;
}

Image Image::operator-(const Image& a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) - a.getValue(i,j));
	}
    }
    return out;
}

Image& Image::operator*=(double a) {
    for (int i = 0; i < npix; i++) {
	this->data[i] *= a;
    }
    return *this;
}

Image& Image::operator*=(const Image& a) {
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] *= a.getValue(i,j);
	}
    }
    return *this;
}

Image Image::operator*(double a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) * a);
	}
    }
    return out;
}

Image Image::operator*(const Image& a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) * a.getValue(i,j));
	}
    }
    return out;
}

Image& Image::operator/=(double a) {
    for (int i = 0; i < npix; i++) {
	this->data[i] /= a;
    }
    return *this;
}

Image& Image::operator/=(const Image& a) {
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    data[i+j*npx] /= a.getValue(i,j);
	}
    }
    return *this;
}

Image Image::operator/(double a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) / a);
	}
    }
    return out;
}

Image Image::operator/(const Image& a) {
    Image out(npx, npy);
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    out.setValue(i, j, getValue(i, j) / a.getValue(i,j));
	}
    }
    return out;
}

double Image::mean() {
    double v = 0.0;
    int n = 0;
    for (int i = 0; i < npix; i++) {
        v += data[i];
        n++;
    }
    return (v/n);
}

double Image::std() {
    double v = 0.0;
    double v2 = 0.0;
    int n = 0;
    for (int i = 0; i < npix; i++) {
        v2 += data[i] * data[i];
        v += data[i];
        n++;
    }
    return (sqrt((v2-v*v/n)/n));
}

double Image::median() {
    std::vector<double> v;
    for (int i = 0; i < npix; i++) {
	v.push_back(data[i]);
    }
    std::sort(v.begin(), v.end());
    return (v[npix/2]);
}

Image Image::boxavg(int width) {
    Image out(npx, npy);
    int wh = width / 2;
    int w2 = width * width;
    for (int j = wh; j < npy-wh; j++) {
	for (int i = wh; i < npx-wh; i++) {
	    double v = 0.0;
	    for (int jj = -wh; jj <= wh; jj++) {
		for (int ii = -wh; ii <= wh; ii++) {
		    v += getValue(i+ii,j+jj)/w2;
		}
	    }
	    out.setValue(i, j, v);
	}
    }
    return out;
}

Image Image::convolve(const Image& a) {
    Image out(npx, npy);
    int width = a.getNpx();
    int wh = width / 2;
    for (int j = wh; j < npy-wh; j++) {
	for (int i = wh; i < npx-wh; i++) {
	    double v = 0.0;
	    for (int jj = -wh; jj <= wh; jj++) {
		for (int ii = -wh; ii <= wh; ii++) {
		    v += getValue(i+ii,j+jj) * a.getValue(wh-ii,wh-jj);
		}
	    }
	    out.setValue(i, j, v);
	}
    }
    return out;
}

std::vector<Point> Image::findPeaks(double thres) {
    std::vector<Point> v;
    for (int j = 1; j < npy-1; j++) {
	for (int i = 1; i < npx-1; i++) {
	    if (getValue(i,j) > thres &&
		getValue(i,j) > getValue(i-1,j) &&
		getValue(i,j) > getValue(i+1,j) &&
		getValue(i,j) > getValue(i,j-1) &&
		getValue(i,j) > getValue(i,j+1)) {
	        v.push_back(Point(i,j,getValue(i,j)));
	    }		
	}
    }
    return v;
}

Image Image::getMedian(std::vector<Image> &v) {
    int npx = v[0].getNpx();
    int npy = v[0].getNpy();
    Image out(npx, npy);
    std::vector<double> a;
    for (int j = 0; j < npy; j++) {
	for (int i = 0; i < npx; i++) {
	    for (int k = 0; k < v.size(); k++) {
		a.push_back(v[k].getValue(i, j));
	    }
	    std::sort(a.begin(), a.end());
	    double med = a[a.size()/2];
	    out.setValue(i, j, med);
	    a.clear();
	}
    }
    return out;
}

Image Image::gaussianImage(double sigma) {
    int np = ((int)floor(sigma * 3) + 1) * 2 + 1;
    int wh = np / 2;
    Image out(np, np);
    for (int j = 0; j < np; j++) {
	int y = j - wh;
	for (int i = 0; i < np; i++) {
	    int x = i - wh;
	    double v = exp(-0.5*(x*x+y*y)/(sigma*sigma))/(2.0*M_PI*sigma*sigma);
	    out.setValue(i, j, v);
	}
    }
    return out;
}
